"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.extractMachines = void 0;
const path_1 = __importDefault(require("path"));
const xstate_1 = require("xstate");
const ts_morph_1 = require("ts-morph");
const extractor_1 = require("./extractor");
let projectCache = new Map();
Array.prototype.flatMap = function (iteratee) {
    return [].concat(...this.map(iteratee));
};
const getTsPropertyNameText = (propertyName) => {
    if (ts_morph_1.ts.isIdentifier(propertyName) ||
        ts_morph_1.ts.isStringLiteral(propertyName) ||
        ts_morph_1.ts.isNumericLiteral(propertyName)) {
        return propertyName.text;
    }
    if (ts_morph_1.ts.isComputedPropertyName(propertyName)) {
        throw new Error("Private identifiers can't be used as property names within config object.");
    }
    if (ts_morph_1.ts.isPrivateIdentifier(propertyName)) {
        throw new Error("Private identifiers can't be used as property names within config object.");
    }
    // propertyName is already never here, but TS doesn't recognize that this explores all possibilities here
    // we have to throw so it doesn't complain about the string return type
    throw new Error('This should be unreachable.');
};
exports.extractMachines = async (filePath) => {
    const resolvedFilePath = path_1.default.resolve(process.cwd(), filePath);
    let configFileName = ts_morph_1.ts.findConfigFile(resolvedFilePath, ts_morph_1.ts.sys.fileExists);
    if (!configFileName) {
        throw new Error('No tsconfig.json file could be found');
    }
    let isFreshProject = false;
    if (!projectCache.has(configFileName)) {
        isFreshProject = true;
        const cachedProject = new ts_morph_1.Project({
            tsConfigFilePath: configFileName,
        });
        projectCache.set(configFileName, cachedProject);
    }
    let project = projectCache.get(configFileName);
    project.addSourceFileAtPath(resolvedFilePath);
    project.resolveSourceFileDependencies();
    if (!isFreshProject) {
        let sourceFiles = project.getSourceFiles();
        for (let sourceFile of sourceFiles) {
            sourceFile.refreshFromFileSystemSync();
        }
    }
    let sourceFile = project.getSourceFileOrThrow(resolvedFilePath);
    const machineReferences = sourceFile
        .getImportDeclarations()
        .filter((importDeclar) => importDeclar.getModuleSpecifierValue() === '@xstate/compiled')
        .flatMap((importDeclar) => {
        if (importDeclar.getNamespaceImport()) {
            throw new Error('Namespace imports are not supported yet.');
        }
        return importDeclar.getNamedImports();
    })
        .filter((namedImport) => namedImport.getNameNode().compilerNode.text === 'createMachine')
        .flatMap((namedImport) => (namedImport.getAliasNode() || namedImport.getNameNode()).findReferences())
        .flatMap((references) => references.getReferences())
        .map((reference) => reference.getNode())
        .filter((referenceNode) => {
        if (referenceNode.getSourceFile() !== sourceFile) {
            return false;
        }
        const statement = referenceNode.getFirstAncestor(ts_morph_1.Node.isStatement);
        return !statement || !ts_morph_1.Node.isImportDeclaration(statement);
    });
    return machineReferences.map((machineReference) => {
        const machineCall = machineReference.getParent();
        if (!ts_morph_1.Node.isCallExpression(machineCall) ||
            machineCall.getExpression() !== machineReference) {
            throw new Error("`Machine` can only be called - you can't pass it around or do anything else with it.");
        }
        const configNode = machineCall.getArguments()[0];
        if (ts_morph_1.Node.isObjectLiteralExpression(configNode)) {
            configNode.transform((traversal) => {
                if (traversal.currentNode === configNode.compilerNode) {
                    return traversal.visitChildren();
                }
                if (ts_morph_1.ts.isPropertyAssignment(traversal.currentNode) &&
                    getTsPropertyNameText(traversal.currentNode.name) !== 'context' &&
                    (ts_morph_1.ts.isObjectLiteralExpression(traversal.currentNode.initializer) ||
                        ts_morph_1.ts.isArrayLiteralExpression(traversal.currentNode.initializer) ||
                        ts_morph_1.ts.isStringLiteral(traversal.currentNode.initializer))) {
                    return ts_morph_1.ts.updatePropertyAssignment(traversal.currentNode, traversal.currentNode.name, ts_morph_1.ts.factory.createAsExpression(traversal.currentNode.initializer, ts_morph_1.ts.factory.createTypeReferenceNode(ts_morph_1.ts.factory.createIdentifier('const'), undefined)));
                }
                return traversal.currentNode;
            });
        }
        const stateSchema = extractor_1.extractSchema(configNode.getType());
        const optionsNode = machineCall.getArguments()[1];
        const options = optionsNode && extractor_1.extractOptions(optionsNode.getType());
        const secondTypeArg = machineCall.getTypeArguments()[2];
        if (!ts_morph_1.Node.isLiteralTypeNode(secondTypeArg)) {
            throw new Error('Second type argument passed to `Machine` has to be a string literal.');
        }
        const literal = secondTypeArg.getLiteral();
        if (!ts_morph_1.Node.isStringLiteral(literal)) {
            throw new Error('Second type argument passed to `Machine` has to be a string literal.');
        }
        return {
            id: literal.getLiteralValue(),
            machine: xstate_1.createMachine(stateSchema, options),
        };
    });
};
