"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.extractOptions = exports.extractSchema = void 0;
const path_1 = __importDefault(require("path"));
const knownActions = new Set([
    'AssignAction',
    'ChooseAction',
    'LogAction',
    'PureAction',
    'RaiseAction',
    'SendAction',
]);
const dummyNamedFunction = () => { };
const findLastIndex = (arr, predicate) => {
    for (let index = arr.length - 1; index >= 0; index--) {
        if (predicate(arr[index])) {
            return index;
        }
    }
    return -1;
};
const explodeAbsolutePath = (filePath) => {
    if (!path_1.default.isAbsolute(filePath)) {
        throw new Error(`Received file path is not an absolute one: ${filePath}`);
    }
    const parsed = path_1.default.parse(filePath);
    return [...parsed.dir.slice(parsed.root.length).split(path_1.default.sep), parsed.base];
};
const isNodeModulePath = (moduleName, absoluteFilePath) => {
    if (!moduleName.length) {
        throw new Error('Received `moduleName` has to be non-empty.');
    }
    const explodedPath = explodeAbsolutePath(absoluteFilePath);
    const nodeModulesIndex = findLastIndex(explodedPath, (segment) => segment === 'node_modules');
    if (nodeModulesIndex === -1) {
        return false;
    }
    if (moduleName[0] === '@') {
        return (`${explodedPath[nodeModulesIndex + 1]}/${explodedPath[nodeModulesIndex + 2]}` === moduleName);
    }
    return explodedPath[nodeModulesIndex + 1] === moduleName;
};
const indexer = Symbol('schema.extractor.indexer');
const lazy = (getter) => ({
    extract: (type) => getter().extract(type),
});
const object = (shape) => ({
    extract(type) {
        if (!(type === null || type === void 0 ? void 0 : type.isObject())) {
            return [true, undefined];
        }
        if (type.getStringIndexType() || type.getNumberIndexType()) {
            return [true, undefined];
        }
        const extracted = {};
        for (const propertySymbol of type.getProperties()) {
            const propName = propertySymbol.getName();
            const propExtractor = shape[propName] || shape[indexer];
            if (!propExtractor) {
                // for now ignore properties not known to the shape
                // TODO: rethink if that's the best solution
                // list pros and cons of the extractor being strict versus it only handling what it knows
                continue;
            }
            const [err, value, hasValue] = propExtractor.extract(propertySymbol.getValueDeclarationOrThrow().getType());
            if (err) {
                return [err, undefined];
            }
            if (hasValue) {
                extracted[propName] = value;
            }
        }
        return [false, extracted, true];
    },
});
const optional = (t) => ({
    extract(type) {
        // TODO: should this accept explicit undefined as well?
        if (!type) {
            return [false, undefined, false];
        }
        return t.extract(type);
    },
});
const array = (typeExtractor) => ({
    extract(type) {
        if (!(type === null || type === void 0 ? void 0 : type.isTuple())) {
            return [true, undefined];
        }
        const result = [];
        for (const element of type.getTupleElements()) {
            const [err, value, hasValue] = typeExtractor.extract(element);
            if (err) {
                return [err, undefined];
            }
            if (hasValue) {
                result.push(value);
            }
        }
        return [false, result, true];
    },
});
const match = (candidates) => ({
    extract(type) {
        for (const candidate of candidates) {
            const [, value, hasValue] = candidate.extract(type);
            if (hasValue) {
                return [false, value, true];
            }
        }
        return [true, undefined];
    },
});
const undef = () => ({
    extract(type) {
        if (!(type === null || type === void 0 ? void 0 : type.isUndefined())) {
            return [true, undefined];
        }
        return [false, undefined, true];
    },
});
const bool = (literal) => ({
    extract(type) {
        if (!(type === null || type === void 0 ? void 0 : type.isBooleanLiteral())) {
            return [true, undefined];
        }
        const actual = type.compilerType.intrinsicName === 'true';
        if (literal !== undefined && literal !== actual) {
            return [true, undefined];
        }
        return [false, actual, true];
    },
});
const string = (literals) => ({
    extract(type) {
        if (!(type === null || type === void 0 ? void 0 : type.isStringLiteral())) {
            return [true, undefined];
        }
        const literal = type.compilerType.value;
        if (!literals || literals.includes(literal)) {
            return [false, literal, true];
        }
        return [true, undefined];
    },
});
const number = () => ({
    extract(type) {
        if (!(type === null || type === void 0 ? void 0 : type.isNumber())) {
            return [true, undefined];
        }
        // provide a dummy value since for numbers we don't need to extract literals right now
        return [false, 10, true];
    },
});
const func = () => ({
    extract(type) {
        if (!(type === null || type === void 0 ? void 0 : type.getCallSignatures().length)) {
            return [true, undefined];
        }
        return [false, dummyNamedFunction, true];
    },
});
const SingleOrArray = (typeExtractor) => match([typeExtractor, array(typeExtractor)]);
const Action = () => ({
    extract(type) {
        if (!(type === null || type === void 0 ? void 0 : type.isObject())) {
            return [true, undefined];
        }
        const symbol = type.getSymbol();
        if (!symbol) {
            return [true, undefined];
        }
        if (!knownActions.has(symbol.getName()) ||
            !symbol
                .getDeclarations()
                .some((declaration) => isNodeModulePath('xstate', declaration.getSourceFile().getFilePath()))) {
            return [true, undefined];
        }
        return [false, dummyNamedFunction, true];
    },
});
const Actions = SingleOrArray(match([string(), func(), Action()]));
const Target = match([undef(), SingleOrArray(string())]);
const Transition = match([
    Target,
    object({
        target: optional(Target),
        cond: optional(match([string(), func()])),
        actions: optional(Actions),
        internal: optional(bool()),
    }),
]);
const TransitionsMap = object({
    [indexer]: SingleOrArray(Transition),
});
const Invoke = SingleOrArray(object({
    // TODO: this can be an object with .type
    src: string(),
    id: optional(string()),
    onDone: optional(SingleOrArray(Transition)),
    onError: optional(SingleOrArray(Transition)),
    autoForward: optional(bool()),
}));
const State = object({
    type: optional(string(['atomic', 'compound', 'parallel', 'final', 'history'])),
    id: optional(string()),
    initial: optional(string()),
    entry: optional(Actions),
    exit: optional(Actions),
    invoke: optional(Invoke),
    on: optional(TransitionsMap),
    after: optional(TransitionsMap),
    always: optional(SingleOrArray(Transition)),
    states: optional(lazy(() => States)),
    onDone: optional(SingleOrArray(Transition)),
    // TODO: supported on final states, implement it
    // data: ?
    history: optional(match([string(['shallow', 'deep']), bool(true)])),
    // XState seems to allow undefined here, that's weird? what would it mean?
    // it also only allows StateValue, need to recheck how the whole thing behaves
    // let's keep this defined as a simple string for now
    target: optional(string()),
});
const States = object({
    [indexer]: State,
});
const Options = optional(object({
    actions: optional(object({
        [indexer]: match([func(), Action()]),
    })),
    delays: optional(object({
        [indexer]: match([func(), number()]),
    })),
    guards: optional(object({
        [indexer]: func(),
    })),
    services: optional(object({
        [indexer]: func(),
    })),
}));
exports.extractSchema = (type) => {
    const [error, schema] = State.extract(type);
    if (error) {
        throw new Error('Could not extract state schema');
    }
    return schema;
};
exports.extractOptions = (type) => {
    const [error, options] = Options.extract(type);
    if (error) {
        throw new Error('Could not extract options');
    }
    return options;
};
