"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.introspectMachine = exports.getMatchesStates = void 0;
require("colors");
const utils_1 = require("xstate/lib/utils");
const traversalUtils_1 = require("./traversalUtils");
exports.getMatchesStates = (machine) => {
    const allStateNodes = machine.stateIds.map((id) => machine.getStateNodeById(id));
    const states = allStateNodes.reduce((arr, node) => {
        return [
            ...arr,
            ...utils_1.toStatePaths(utils_1.pathToStateValue(node.path)).map((path) => path.join('.')),
        ];
    }, []);
    return states;
};
const toCompactArray = (maybeArray) => {
    if (!maybeArray) {
        return [];
    }
    if (Array.isArray(maybeArray)) {
        return maybeArray;
    }
    return [maybeArray];
};
const getStringActions = (_actions) => toCompactArray(_actions).filter((action) => typeof action === 'string');
const makeSubStateFromNode = (node, rootNode, nodeMaps) => {
    const nodeFromMap = nodeMaps[node.id];
    const stateNode = rootNode.getStateNodeById(node.id);
    const targets = traversalUtils_1.getTransitionsFromNode(stateNode);
    return {
        sources: Array.from(nodeFromMap.sources)
            .filter(Boolean)
            .map((event) => `'${event}'`)
            .join(' | ') || 'never',
        targets: Array.from(targets)
            .filter(Boolean)
            .map((event) => `'${event}'`)
            .join(' | ') || 'never',
        states: Array.from(nodeFromMap.children).reduce((obj, child) => {
            const childNode = rootNode.getStateNodeById(child);
            return Object.assign(Object.assign({}, obj), { [childNode.key]: makeSubStateFromNode(childNode, rootNode, nodeMaps) });
        }, {}),
    };
};
class ItemMap {
    constructor(props) {
        /**
         * The internal map that we use to keep track
         * of all of the items
         */
        this.map = {};
        this.checkIfOptional = props.checkIfOptional;
    }
    /**
     * Add an item to the cache, along with the path of the node
     * it occurs on
     */
    addItem(itemName, nodePath) {
        if (!this.map[itemName]) {
            this.map[itemName] = {
                events: new Set(),
                states: new Set(),
            };
        }
        this.map[itemName].states.add(utils_1.pathToStateValue(nodePath));
    }
    /**
     * Add a triggering event to an item in the cache, for
     * instance the event type which triggers a guard/action/service
     */
    addEventToItem(itemName, eventType, nodePath) {
        this.addItem(itemName, nodePath);
        this.map[itemName].events.add(eventType);
    }
    /**
     * Transform the data into the shape required for index.d.ts
     */
    toDataShape() {
        let isRequiredInTotal = false;
        const lines = Object.entries(this.map)
            .filter(([name]) => {
            return !/\./.test(name);
        })
            .map(([name, data]) => {
            const optional = this.checkIfOptional(name);
            if (!optional) {
                isRequiredInTotal = true;
            }
            return {
                name,
                required: !optional,
                events: Array.from(data.events).filter(Boolean),
                states: Array.from(data.states)
                    .map((state) => JSON.stringify(state))
                    .filter(Boolean),
            };
        });
        return {
            lines,
            required: isRequiredInTotal,
        };
    }
}
exports.introspectMachine = (machine) => {
    const guards = new ItemMap({
        checkIfOptional: (name) => Boolean(machine.options.guards[name]),
    });
    const actions = new ItemMap({
        checkIfOptional: (name) => Boolean(machine.options.actions[name]),
    });
    const services = new ItemMap({
        checkIfOptional: (name) => Boolean(machine.options.services[name]),
    });
    const activities = new ItemMap({
        checkIfOptional: (name) => Boolean(machine.options.activities[name]),
    });
    const delays = new ItemMap({
        checkIfOptional: (name) => Boolean(machine.options.delays[name]),
    });
    const nodeMaps = {};
    const allStateNodes = machine.stateIds.map((id) => machine.getStateNodeById(id));
    allStateNodes.forEach((node) => {
        nodeMaps[node.id] = {
            sources: new Set(),
            children: new Set(),
        };
    });
    allStateNodes.forEach((node) => {
        var _a, _b, _c, _d, _e;
        (_a = Object.values(node.states)) === null || _a === void 0 ? void 0 : _a.forEach((childNode) => {
            nodeMaps[node.id].children.add(childNode.id);
        });
        // TODO - make activities pick up the events
        // that led to them
        (_b = node.activities) === null || _b === void 0 ? void 0 : _b.forEach((activity) => {
            if (/\./.test(activity.type))
                return;
            if (activity.type && activity.type !== 'xstate.invoke') {
                activities.addItem(activity.type, node.path);
            }
        });
        (_c = node.after) === null || _c === void 0 ? void 0 : _c.forEach(({ delay }) => {
            if (typeof delay === 'string') {
                delays.addItem(delay, node.path);
            }
        });
        (_d = node.invoke) === null || _d === void 0 ? void 0 : _d.forEach((service) => {
            if (typeof service.src !== 'string' || /\./.test(service.src))
                return;
            services.addItem(service.src, node.path);
        });
        (_e = node.transitions) === null || _e === void 0 ? void 0 : _e.forEach((transition) => {
            var _a, _b;
            (_a = transition.target) === null || _a === void 0 ? void 0 : _a.forEach((targetNode) => {
                nodeMaps[targetNode.id].sources.add(transition.eventType);
            });
            if (transition.cond && transition.cond.name) {
                if (transition.cond.name !== 'cond') {
                    guards.addEventToItem(transition.cond.name, transition.eventType, node.path);
                }
            }
            (_b = transition.target) === null || _b === void 0 ? void 0 : _b.forEach((targetNode) => {
                var _a;
                /** Pick up invokes */
                (_a = targetNode.invoke) === null || _a === void 0 ? void 0 : _a.forEach((service) => {
                    if (typeof service.src !== 'string' || /\./.test(service.src))
                        return;
                    services.addEventToItem(service.src, transition.eventType, node.path);
                });
            });
            transition.actions
                .filter((action) => !/^xstate\./.test(action.type))
                .forEach((action) => {
                actions.addEventToItem(action.type, transition.eventType, node.path);
            });
        });
    });
    allStateNodes.forEach((node) => {
        const allActions = [];
        const stringEntryActions = getStringActions(node.config.entry);
        const stringExitActions = getStringActions(node.config.exit);
        allActions.push(...stringEntryActions);
        allActions.push(...stringExitActions);
        allActions.forEach((action) => {
            actions.addItem(action, node.path);
        });
        stringEntryActions.forEach((action) => {
            const sources = nodeMaps[node.id].sources;
            sources === null || sources === void 0 ? void 0 : sources.forEach((source) => {
                actions.addEventToItem(action, source, node.path);
            });
        });
    });
    const subState = makeSubStateFromNode(machine, machine, nodeMaps);
    return {
        stateMatches: exports.getMatchesStates(machine),
        subState,
        guards: guards.toDataShape(),
        actions: actions.toDataShape(),
        services: services.toDataShape(),
        activities: activities.toDataShape(),
        delays: delays.toDataShape(),
    };
};
