"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.printJsFiles = exports.printToFile = exports.getNodeModulesDir = exports.getDeclarationFileTexts = void 0;
require("colors");
const fs_1 = __importDefault(require("fs"));
const handlebars_1 = __importDefault(require("handlebars"));
const handlebars_helpers_1 = __importDefault(require("handlebars-helpers"));
const path_1 = __importDefault(require("path"));
const pkg_up_1 = __importDefault(require("pkg-up"));
const rimraf_1 = __importDefault(require("rimraf"));
const ensureFolderExists = (absoluteDir) => {
    if (fs_1.default.existsSync(absoluteDir)) {
        return;
    }
    fs_1.default.mkdirSync(absoluteDir);
};
const ensureMultipleFoldersExist = (absoluteRoot, paths) => {
    let concatenatedPath = absoluteRoot;
    paths.forEach((dir) => {
        concatenatedPath = path_1.default.join(concatenatedPath, dir);
        ensureFolderExists(concatenatedPath);
    });
};
const renderSubstate = (subState) => {
    return `{
    targets: ${subState.targets};
    sources: ${subState.sources};
    states: {
      ${Object.entries(subState.states)
        .map(([key, state]) => {
        return `${key}: ${renderSubstate(state)}`;
    })
        .join('\n')}
    };
  }`;
};
exports.getDeclarationFileTexts = (cache) => {
    const indexTemplateString = fs_1.default
        .readFileSync(path_1.default.resolve(__dirname, './templates/index.d.ts.hbs'))
        .toString();
    const reactTemplateString = fs_1.default
        .readFileSync(path_1.default.resolve(__dirname, './templates/react.d.ts.hbs'))
        .toString();
    handlebars_helpers_1.default({
        handlebars: handlebars_1.default,
    });
    const indexTemplate = handlebars_1.default.compile(indexTemplateString);
    const reactTemplate = handlebars_1.default.compile(reactTemplateString);
    const machines = Object.values(cache).map((machine) => ({
        id: machine.id,
        machine: Object.assign(Object.assign({}, machine), { subState: renderSubstate(machine.subState) }),
    }));
    return {
        'index.d.ts': indexTemplate({ machines }),
        'react.d.ts': reactTemplate({ machines }),
    };
};
exports.getNodeModulesDir = () => {
    const packageJson = pkg_up_1.default.sync();
    if (!packageJson) {
        throw new Error('Could not find a package.json in any directory in or above this one.');
    }
    const targetDir = path_1.default.resolve(path_1.default.dirname(packageJson), 'node_modules');
    return targetDir;
};
exports.printToFile = (cache, outDir) => {
    const files = exports.getDeclarationFileTexts(cache);
    const nodeModulesDir = exports.getNodeModulesDir();
    const targetDir = path_1.default.resolve(nodeModulesDir, '@xstate/compiled');
    /** Delete @xstate/compiled directory so that it triggers VSCode to re-check it */
    rimraf_1.default.sync(path_1.default.join(targetDir, '*.d.ts'));
    exports.printJsFiles();
    ensureMultipleFoldersExist(nodeModulesDir, ['@xstate', 'compiled']);
    fs_1.default.writeFileSync(outDir
        ? path_1.default.resolve(process.cwd(), outDir, 'index.d.ts')
        : path_1.default.join(targetDir, 'index.d.ts'), files['index.d.ts']);
    fs_1.default.writeFileSync(outDir
        ? path_1.default.resolve(process.cwd(), outDir, 'react.d.ts')
        : path_1.default.join(targetDir, 'react.d.ts'), files['react.d.ts']);
    if (outDir) {
        // If the user specifies an outDir, we need to add some dummy types
        // so that we can override something
        fs_1.default.writeFileSync(path_1.default.join(targetDir, 'react.d.ts'), `export default any;`);
        fs_1.default.writeFileSync(path_1.default.join(targetDir, 'index.d.ts'), `export default any;`);
    }
};
/**
 * Prints the js files, which needs to be done in advance
 * of rollup looking at the code to ensure there is a module for rollup
 * to statically analyse
 */
exports.printJsFiles = () => {
    const nodeModulesDir = exports.getNodeModulesDir();
    const targetDir = path_1.default.resolve(nodeModulesDir, '@xstate/compiled');
    ensureMultipleFoldersExist(nodeModulesDir, ['@xstate', 'compiled']);
    const indexJsTemplate = fs_1.default
        .readFileSync(path_1.default.resolve(__dirname, './templates/index.js.hbs'))
        .toString();
    const indexEsTemplate = fs_1.default
        .readFileSync(path_1.default.resolve(__dirname, './templates/index.es.js.hbs'))
        .toString();
    const reactJsTemplate = fs_1.default
        .readFileSync(path_1.default.resolve(__dirname, './templates/react.js.hbs'))
        .toString();
    const packageJsonTemplate = fs_1.default
        .readFileSync(path_1.default.resolve(__dirname, './templates/package.json.hbs'))
        .toString();
    fs_1.default.writeFileSync(path_1.default.join(targetDir, 'index.js'), indexJsTemplate);
    fs_1.default.writeFileSync(path_1.default.join(targetDir, 'index.es.js'), indexEsTemplate);
    fs_1.default.writeFileSync(path_1.default.join(targetDir, 'react.js'), reactJsTemplate);
    fs_1.default.writeFileSync(path_1.default.join(targetDir, 'package.json'), packageJsonTemplate);
};
